# Author: Mark Blakeney, Feb 2024.
'''
Install Python applications into isolated virtual environments and
create links to the executables in a bin directory for your PATH. Like
pipx but uses uv instead of venv + pip.
'''
from __future__ import annotations

import argparse
import importlib
import os
import re
import shlex
import sys
from pathlib import Path
from typing import Optional

import platformdirs

from . import utils
from .run import run

DEFUV = 'uv'
DEFPY = 'python3'

# Some constants
MOD = Path(__file__)
PROG = MOD.stem
PROGU = PROG.upper()
CNFFILE = platformdirs.user_config_path(f'{PROG}-flags.conf')

def path_check(bin_name: str, bin_dir: Path) -> str:
    'Check and report that users PATH is set up correctly'
    path = os.getenv('PATH')
    if not path:
        return 'WARNING: Your PATH is not set.'

    if str(bin_dir) not in path.split(':'):
        return f'WARNING: Your PATH does not contain {bin_name} ({bin_dir}).'

    return f'Your PATH contains {bin_name} ({bin_dir}).'

def main() -> Optional[str]:
    'Main code'
    mainparser = argparse.ArgumentParser(description=__doc__,
        epilog='Note you can set default starting global options '
        f'in {CNFFILE}.')
    mainparser.add_argument('--uv', metavar='uv_path',
                            help=f'path to uv executable, default="{DEFUV}"')
    mainparser.add_argument('-m', '--no-man-pages', action='store_true',
                            help='do not install package man pages')
    mainparser.add_argument('--home',
                            help=f'specify {PROGU}_HOME')
    mainparser.add_argument('--bin-dir',
                            help=f'specify {PROGU}_BIN_DIR')
    mainparser.add_argument('--man-dir',
                            help=f'specify {PROGU}_MAN_DIR')
    mainparser.add_argument('--default-python',
                            help='path to default python executable, '
                            f'default="{DEFPY}"')
    mainparser.add_argument('-V', '--version', action='store_true',
                            help=f'just print {PROG} version and exit')
    subparser = mainparser.add_subparsers(title='Commands',
            dest='func')

    # Iterate over the commands to set up their parsers
    progs = {}
    for modfile in sorted((MOD.parent / 'commands').glob('[!_]*.py')):
        name = modfile.stem
        mod = importlib.import_module(f'{PROG}.commands.{name}')
        docstr = mod.__doc__.strip().split('\n\n')[0] if mod.__doc__ else None
        parser = subparser.add_parser(name, description=mod.__doc__,
                formatter_class=argparse.RawDescriptionHelpFormatter,
                help=docstr)

        progs[name] = parser.prog

        if hasattr(mod, 'init'):
            mod.init(parser)

        if not hasattr(mod, 'main'):
            mainparser.error(f'"{name}" command must define a main()')

        parser.set_defaults(func=mod.main, parser=parser, name=name)

    # Merge in default args from user config file. Then parse the
    # command line.
    if CNFFILE.exists():
        with CNFFILE.open() as fp:
            lines = [re.sub(r'#.*$', '', line).strip() for line in fp]
        cnflines = ' '.join(lines).strip()
    else:
        cnflines = ''

    args = mainparser.parse_args(shlex.split(cnflines) + sys.argv[1:])

    if args.version:
        print(f'{PROG}=={utils.version()}')
        return None

    homedir = args.home or os.getenv(f'{PROGU}_HOME')
    bindir = args.bin_dir or os.getenv(f'{PROGU}_BIN_DIR')
    mandir = args.man_dir or os.getenv(f'{PROGU}_MAN_DIR')
    pyexe = args.default_python or os.getenv(f'{PROGU}_DEFAULT_PYTHON') or DEFPY

    if os.geteuid() == 0:
        home_dir = Path(homedir if homedir else f'/opt/{PROG}')
        bin_dir = Path(bindir if bindir else '/usr/local/bin')
        man_dir = Path(mandir if mandir else '/usr/local/share/man')
    else:
        home_dir = Path(homedir) if homedir else \
                        (platformdirs.user_data_path() / PROG)
        bin_dir = Path(bindir) if bindir else \
                Path('~/.local/bin').expanduser()
        man_dir = Path(mandir) if mandir else \
                Path('~/.local/share/man').expanduser()

    if not args.func:
        mainparser.print_help()
        print('\nEnvironment:')
        print(f'{PROGU}_HOME = {home_dir}')
        print(f'{PROGU}_BIN_DIR = {bin_dir}')
        print(f'{PROGU}_MAN_DIR = {man_dir}')
        print(f'{PROGU}_DEFAULT_PYTHON = {pyexe}')
        print()
        print(path_check(f'{PROGU}_BIN_DIR', bin_dir))
        return None

    # Ensure uv is installed/available
    uv = args.uv or DEFUV
    version = run(f'{uv} --version', capture=True)
    if not version:
        if args.uv:
            return f'Error: uv program not found at "{uv}"'

        return f'Error: {uv} program must be installed, and in your PATH '\
                'or specified with --uv option.'

    # Keep some useful info in the namespace passed to the command
    args._uv = uv
    args._packages_dir = home_dir / 'packages'
    args._packages_dir.mkdir(parents=True, exist_ok=True)
    args._venvs_dir = home_dir / 'venvs'
    args._venvs_dir.mkdir(parents=True, exist_ok=True)
    args._bin_dir = bin_dir
    args._bin_dir.mkdir(parents=True, exist_ok=True)
    args._man_dir = man_dir
    if not args.no_man_pages:
        args._man_dir.mkdir(parents=True, exist_ok=True)
    args._pyexe = pyexe
    args._prog = PROG
    args._meta_file = f'{PROG}_metadata.json'
    args._freeze_file = f'{PROG}_freeze.txt'
    if not hasattr(args, 'verbose'):
        args.verbose = False

    # Purge any old files left lying around
    utils.purge_old_files(args)

    # Run the command that the user specified
    return args.func(args)

if __name__ == '__main__':
    sys.exit(main())



# Author: Mark Blakeney, Feb 2024.
'Install a Python application using an isolated virtual environment.'
from __future__ import annotations

from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Optional

from filelock import FileLock
from platformdirs import user_runtime_path

from .. import utils
from ..run import run

MAX_VDIRS = 1_000_000

def get_next_vdir(vdirbase: Path) -> Optional[Path]:
    'Return the first available venv directory'
    vdirs = set(int(f.name) for f in vdirbase.iterdir())
    for n in range(1, MAX_VDIRS + 1):
        if n not in vdirs:
            return vdirbase / str(n)

    return None

def init(parser: ArgumentParser) -> None:
    'Called to add command arguments to parser at init'
    xgroup = parser.add_mutually_exclusive_group()
    xgroup.add_argument('-p', '--python',
                        help='specify explicit python executable path')
    xgroup.add_argument('-P', '--pyenv',
                        help='pyenv python version to use, '
                        'i.e. from `pyenv versions`, e.g. "3.9".')
    parser.add_argument('-f', '--force', action='store_true',
                        help='recreate any existing venv')
    parser.add_argument('-e', '--editable', action='store_true',
                        help='install application[s] in editable mode')
    parser.add_argument('-d', '--include-deps', action='store_true',
                        help='include executables from dependencies')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='give more output')
    parser.add_argument('package', nargs='+',
                        help='application[s] to install')

def main(args: Namespace) -> Optional[str]:
    'Called to action this command'
    pyexe = utils.get_python(args)
    venv_args = utils.make_args((args.verbose, '-v'), (not args.verbose, '-q'),
                                (bool(pyexe), f'--python={pyexe}'))
    pip_args = utils.make_args((args.verbose, '-v'), (args.editable, '-e'))

    lockfile = user_runtime_path() / f'{args._prog}.lock'
    vdirbase = args._venvs_dir
    for pkg in args.package:
        # Use a lock file in case we are running multiple installs in parallel
        with FileLock(lockfile):
            vdir = get_next_vdir(vdirbase)
            if not vdir:
                return f'Error: Too many vdirs (>{MAX_VDIRS}) in {vdirbase}'

            # Create the vdir
            if not run(f'uv venv{venv_args} {vdir}'):
                utils.rm_vdir(vdir, args)
                return f'Error: failed to create {vdir} for {pkg}.'

        python_exe = (vdir / 'bin' / 'python').resolve()
        python_ver = run(f'{python_exe} -V', capture=True)
        python_ver = python_ver.strip().split()[1] if python_ver else '?ver?'
        print(f'Created {vdir} using {python_exe} ({python_ver})')

        # Install the package
        if not utils.piprun(vdir, f'install --no-deps{pip_args} "{pkg}"'):
            utils.rm_vdir(vdir, args)
            return f'Error: failed to preinstall "{pkg}".'

        versions = utils.get_versions(vdir)
        if not versions:
            utils.rm_vdir(vdir, args)
            return f'Error: failed to get versions for {pkg}.'

        if len(versions) != 1:
            return f'Error: multiple packages qualified: {list(versions)}'

        pkgname, (vers, editpath) = versions.popitem()
        pdir = Path(args._packages_dir, pkgname)

        if pdir.exists():
            if not args.force:
                utils.rm_vdir(vdir, args)
                return f'Error: venv for {pkgname} exists. Use -f to force.'
            print(f'Removing pre-existing {pkgname} venv dir.')
            utils.rm_vdir(pdir, args)
            pdir.unlink()

        if not utils.piprun(vdir, f'install{pip_args} "{pkg}"'):
            utils.rm_vdir(vdir, args)
            return f'Error: failed to install "{pkg}".'

        pdir.symlink_to(vdir)

        data: dict = {'name': pkgname}
        if editpath:
            data['editpath'] = editpath

        if args.include_deps:
            data['deps'] = True

        err = utils.make_links(vdir, pkgname, args, data)
        if err:
            pdir.unlink()
            utils.rm_vdir(vdir, args)
            return err

    return None



# Author: Mark Blakeney, Feb 2024.
'Upgrade all applications.'
from __future__ import annotations

from argparse import ArgumentParser, Namespace
from typing import Optional

from .. import utils
from . import upgrade

def init(parser: ArgumentParser) -> None:
    'Called to add command arguments to parser at init'
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='give more output')
    parser.add_argument('-s', '--skip', nargs='*',
                        help='skip these applications, e.g. '
                        '"-s package1 package2"')

def main(args: Namespace) -> Optional[str]:
    'Called to action this command'
    args.package = utils.get_all_package_names(args)
    return upgrade.main(args)



# Author: Mark Blakeney, Feb 2024.
'Uninstall an application.'
from __future__ import annotations

from argparse import ArgumentParser, Namespace
from typing import Optional

from .. import utils

def init(parser: ArgumentParser) -> None:
    'Called to add command arguments to parser at init'
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='give more output')
    parser.add_argument('package', nargs='+',
                        help='application[s] to uninstall')

def main(args: Namespace) -> Optional[str]:
    'Called to action this command'
    for pkgname in args.package:
        pkgname, vdir = utils.get_package_from_arg(pkgname, args)
        if not vdir or not utils.rm_package(pkgname, args):
            return f'Application {pkgname} is not installed.'

        print(f'{pkgname} uninstalled.')

    return None



# Author: Mark Blakeney, Feb 2024.
'Common module containing shared functions'
from __future__ import annotations

import subprocess
import sys
from typing import Optional

def run(cmd: str, *, capture: bool = False,
        quiet: bool = False) -> Optional[str]:
    'Run given command string'
    if capture:
        stdout = subprocess.PIPE
    else:
        stdout = None
        if not quiet:
            print(f'>>> Running {cmd}')
    try:
        res = subprocess.run(cmd, shell=True, stdout=stdout,
                             universal_newlines=True)
    except Exception as e:
        print(f'{cmd} failed: {e}', file=sys.stderr)
        return None

    if res.returncode != 0:
        return None

    return (res.stdout and res.stdout.strip()) if capture else 'ok'



# Author: Mark Blakeney, Feb 2024.
'Common module containing shared functions'
from __future__ import annotations

import json
import os
import shutil
import sys
from argparse import Namespace
from pathlib import Path
from typing import Iterable, Optional

from .run import run

def get_json(vdir: Path, args: Namespace) -> Optional[dict]:
    'Get JSON data for this virtual environment'
    tgt = vdir.resolve() / args._meta_file
    try:
        with tgt.open() as fp:
            return json.load(fp)
    except Exception:
        return None

def _set_json(vdir: Path, args: Namespace, data: dict) -> Optional[str]:
    'Set JSON data for this virtual environment'
    tgt = vdir.resolve() / args._meta_file
    try:
        with tgt.open('w') as fp:
            json.dump(data, fp)
    except Exception as e:
        return str(e)

    return None

def piprun(vdir: Path, cmd: str, **args) -> Optional[str]:
    'Run given pip command in the virtual environment'
    os.environ['VIRTUAL_ENV'] = str(vdir.resolve())
    return run(f'uv pip {cmd}', **args)

def get_versions(vdir: Path) -> Optional[dict[str, tuple[str, Optional[str]]]]:
    'Return the versions of the packages in the virtual environment'
    out = piprun(vdir, 'list', capture=True)
    if not out:
        return None

    found = False
    data: dict[str, tuple[str, Optional[str]]] = {}
    for line in out.splitlines():
        if line.startswith('-'):
            found = True
        elif found:
            fields = line.split(maxsplit=2)
            if len(fields) == 2:
                data[fields[0]] = fields[1], None
            else:
                data[fields[0]] = fields[1], fields[2]

    return data

def make_args(*args: tuple[bool, str]) -> str:
    'Build a string of args based on (bool, arg) pairs'
    argstr = ' '.join(v2 for v1, v2 in args if v1)
    return ' ' + argstr if argstr else ''

def _load_record(rfile: Path) -> Iterable[str]:
    'Yield the executable names from a RECORD file'
    with rfile.open() as fp:
        for line in fp:
            line = line.strip()
            if line.startswith('../../../bin/'):
                yield Path(line.split(',', 1)[0]).name

def _link_exists(path: Path) -> bool:
    'Return True if the path is a link (regardless if the target exists)'
    # Equivalent to path.exists(follow_symlinks=False) on Python < 3.12.
    try:
        exists = path.lstat()
    except Exception:
        return False

    return bool(exists)

def _link_app_files(vdir: Path, tgtdir: Path, pkgname: str,
                    args: Namespace, include_deps: bool) -> Iterable[str]:
    'Link app files from entry_points to tgtdir'
    pkgname = pkgname.replace('-', '_')
    for efile in vdir.glob('**/*.dist-info/RECORD'):
        if include_deps or efile.parent.name.startswith(f'{pkgname}-'):
            for app in _load_record(efile):
                srcfile = vdir / 'bin' / app
                if srcfile.is_file() and not srcfile.is_symlink() and \
                      (srcfile.stat().st_mode & 0o111) == 0o111:
                    tgtfile = tgtdir / app
                    if _link_exists(tgtfile):
                        tgtfile.unlink()

                    if args.verbose:
                        print(f'Linking {srcfile} -> {tgtfile}')

                    tgtfile.parent.mkdir(parents=True, exist_ok=True)
                    tgtfile.symlink_to(srcfile)
                    yield srcfile.name

def _link_all_files(srcdir: Path, tgtdir: Path, pat: str,
                    verbose: bool) -> None:
    'Link files from srcdir to tgtdir'
    for srcfile in srcdir.glob(pat):
        tgtfile = tgtdir / srcfile.relative_to(srcdir)
        if _link_exists(tgtfile):
            tgtfile.unlink()

        if verbose:
            print(f'Linking {srcfile} -> {tgtfile}')

        tgtfile.parent.mkdir(parents=True, exist_ok=True)
        tgtfile.symlink_to(srcfile)

def _unlink_all_files(vdir: Path, args: Namespace) -> None:
    'Unlink all link files'
    dirlist = [(args._bin_dir, '*')]
    if args._man_dir.is_dir():
        dirlist.append((args._man_dir, '*/*'))
    for srcdir, pat in dirlist:
        if srcdir.is_dir():
            for file in srcdir.glob(pat):
                if file.is_symlink() and vdir in file.resolve().parents:
                    if args.verbose:
                        print(f'Removing link {file}')
                    file.unlink()

def make_links(vdir: Path, pkgname: str, args: Namespace,
               data: Optional[dict]) -> Optional[str]:
    '[Re-]create app and man page links for package'

    vdir = vdir.resolve()

    # Unlink any existing links
    _unlink_all_files(vdir, args)

    # Link the package applications
    data = data or get_json(vdir, args)
    if not data:
        return 'Error: No JSON data found.'

    include_deps = bool(args.include_deps if hasattr(args, 'include_deps')
                        else data.get('deps'))

    apps = list(_link_app_files(vdir, args._bin_dir, pkgname, args,
                                include_deps))

    # Link all the man pages
    if not args.no_man_pages:
        _link_all_files(vdir / 'share' / 'man', args._man_dir, '*/*',
                        args.verbose)

    # Save the apps in the JSON data
    if not apps:
        return f'Error: {pkgname} has no executables to install.'

    freeze = piprun(vdir, 'freeze', capture=True)
    if not freeze:
        return 'Error: Failed to fetch freeze list.'

    (vdir / args._freeze_file).write_text(freeze)
    data['apps'] = sorted(apps)
    return _set_json(vdir, args, data)

def rm_vdir(vdir: Path, args: Namespace) -> None:
    'Remove all links that point into the virtual environment'
    vdir = vdir.resolve()
    _unlink_all_files(vdir, args)

    # Remove the venv
    if args.verbose:
        print(f'Removing {vdir}')

    shutil.rmtree(vdir)

def add_or_remove_pkg(vdir: Path, pkgname: str, pkgs: list[str],
                      args: Namespace, *, add: bool) -> Optional[str]:
    'Record the addition/removal of a package into the virtual environment'
    data = get_json(vdir, args)
    if not data:
        return 'No JSON data found'

    dataset = set(data.get('injected') or [])
    pkgset = set(pkgs)
    newset = (dataset | pkgset) if add else (dataset - pkgset)

    if newset:
        data['injected'] = sorted(list(newset))
    else:
        data.pop('injected', None)

    return make_links(vdir, pkgname, args, data)

def rm_package(pkgname: str, args: Namespace) -> bool:
    'Remove this package'
    pdir = args._packages_dir / pkgname
    if not pdir.exists():
        return False

    vdir = pdir.resolve()
    pdir.unlink()
    rm_vdir(vdir, args)
    return True

def get_all_pkg_venvs(args: Namespace) -> Iterable[tuple[Path, dict]]:
    'Return a list of all virtual environments and their JSON data'
    for pdir in sorted(args._packages_dir.iterdir()):
        data = get_json(pdir, args)
        if data:
            yield pdir, data

def get_package_from_arg(name: str, args: Namespace) \
        -> tuple[str, Optional[Path]]:
    'Return the package name + vdir corresponding to the given arg, if any'
    if name == '.' or os.sep in name:
        pathstr = str(Path(name).resolve())
        for pdir, data in get_all_pkg_venvs(args):
            if data and data.get('editpath') == pathstr:
                name = pdir.name
                break
        else:
            return name, None

    path = (args._packages_dir / name).resolve()
    return name, (path if path.exists() else None)

def rm_path(path: Path) -> None:
    'Remove the given path'
    print(f'Purging stray {path}', file=sys.stderr)
    if path.is_dir():
        shutil.rmtree(path)
    else:
        path.unlink()

def purge_old_files(args: Namespace) -> None:
    'Clean out any old virtual environments, packages, and executables'
    # Remove any packages that do not point to a dir in the venvs directory
    valids_venvs = set()
    for pkg in args._packages_dir.iterdir():
        vdir = pkg.resolve()
        if vdir.parent != args._venvs_dir or not vdir.is_dir() \
                or not vdir.name.isdigit():
            rm_path(pkg)
        else:
            valids_venvs.add(vdir.name)

    # Remove any venvs that are not in the packages directory
    for vdir in args._venvs_dir.iterdir():
        if vdir.name not in valids_venvs:
            rm_path(vdir)

    # Remove any executables that point to a non-existent path
    for exe in args._bin_dir.iterdir():
        rexe = exe.resolve()
        if args._venvs_dir in rexe.parents and not rexe.exists():
            rm_path(exe)

def get_all_package_names(args: Namespace) -> list[str]:
    'Return a sorted list of all package names'
    return sorted(set(f.name for f in args._packages_dir.iterdir())
                  - set(args.skip or []))

def get_python(args: Namespace) -> str:
    'Return the python executable based on command line args'
    if args.pyenv:
        pyenv_root = run('pyenv root', capture=True)
        if not pyenv_root:
            sys.exit('Error: Can not find pyenv. Is it installed?')

        pyenv_version = run(f'pyenv latest {args.pyenv}', capture=True)
        if not pyenv_version:
            sys.exit(f'Error: no pyenv version {args.pyenv} installed.')

        pyexe = Path(pyenv_root, 'versions', pyenv_version, 'bin', 'python')
        if not pyexe.exists():
            sys.exit(f'Can not determine pyenv version for {args.pyenv}')
    elif args.python:
        pyexe = Path(args.python)
        if not pyexe.exists():
            sys.exit(f'{pyexe} does not exist.')
    else:
        pyexe = args._pyexe

    return str(pyexe)

def version() -> str:
    'Return the version of this package'
    if sys.version_info >= (3, 8):
        from importlib.metadata import version
    else:
        from importlib_metadata import version

    try:
        ver = version(Path(sys.argv[0]).stem)
    except Exception:
        ver = 'unknown'

    return ver



# Author: Mark Blakeney, Feb 2024.
'Upgrade an application.'
from __future__ import annotations

from argparse import ArgumentParser, Namespace
from typing import Optional

from .. import utils

def init(parser: ArgumentParser) -> None:
    'Called to add command arguments to parser at init'
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='give more output')
    parser.add_argument('package', nargs='+',
                        help='application[s] to upgrade')

def main(args: Namespace) -> Optional[str]:
    'Called to action this command'
    pip_args = utils.make_args((args.verbose, '-v'))
    for pkgname in args.package:
        pkgname, vdir = utils.get_package_from_arg(pkgname, args)
        if not vdir:
            return f'Application {pkgname} is not installed.'

        data = utils.get_json(vdir, args) or {}
        editpath = data.get('editpath')
        pkg = f'-e {editpath}' if editpath else pkgname
        extras = ' '.join(data.get('injected', []))
        if not utils.piprun(vdir, f'install --reinstall -U'
                            f'{pip_args} {pkg} {extras}'):
            return f'Error: failed to {args.name} {pkgname}'

        err = utils.make_links(vdir, pkgname, args, data)
        if err:
            return err

        print(f'{pkgname} upgraded.')

    return None



# Author: Mark Blakeney, Feb 2024.
'List all applications installed by this tool.'
from __future__ import annotations

import json
from argparse import ArgumentParser, Namespace
from typing import Optional

from .. import utils

def show(value: str) -> str:
    'Show a string with quotes'
    return f'"{value}"' if isinstance(value, str) else value

def init(parser: ArgumentParser) -> None:
    'Called to add command arguments to parser at init'
    parser.add_argument('--json', action='store_true',
                        help='output json instead')
    parser.add_argument('package', nargs='*',
                        help='list the given application[s] only')

def main(args: Namespace) -> Optional[str]:
    'Called to action this command'
    pkgs = set(utils.get_package_from_arg(p, args)[0] for p in args.package)
    json_out = {}
    for pdir in sorted(args._packages_dir.iterdir()):
        pkgname = pdir.name
        if pkgs and pkgname not in pkgs:
            continue

        data = utils.get_json(pdir, args)
        if data:
            data.pop('name', None)
            if args.json:
                json_out[pkgname] = data
            else:
                d = ', '.join(f'{k}={show(data[k])}' for k in sorted(data))
                print(f'{pkgname}: {d}')

    if args.json:
        print(json.dumps(json_out, indent=2))

    return None



# Author: Mark Blakeney, Feb 2024.
'Reinstall an application.'
from __future__ import annotations

import shutil
import tempfile
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Optional

from .. import utils
from ..run import run

def init(parser: ArgumentParser) -> None:
    'Called to add command arguments to parser at init'
    xgroup = parser.add_mutually_exclusive_group()
    xgroup.add_argument('-p', '--python',
                        help='specify explicit python executable path')
    xgroup.add_argument('-P', '--pyenv',
                        help='pyenv python version to use, '
                        'i.e. from `pyenv versions`, e.g. "3.9".')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='give more output')
    parser.add_argument('package', nargs='+',
                        help='application[s] to reinstall')

def main(args: Namespace) -> Optional[str]:
    'Called to action this command'
    pyexe = utils.get_python(args)
    venv_args = utils.make_args((args.verbose, '-v'), (not args.verbose, '-q'),
                                (bool(pyexe), f'--python={pyexe}'))
    pip_args = utils.make_args((args.verbose, '-v'))
    for pkgname in args.package:
        pkgname, vdir = utils.get_package_from_arg(pkgname, args)
        if not vdir:
            return f'Application {pkgname} is not installed.'

        data = utils.get_json(vdir, args) or {}

        with tempfile.TemporaryDirectory() as tdir:
            tfile = Path(tdir, args._freeze_file)
            shutil.copyfile(vdir / args._freeze_file, tfile)

            # Recreate the vdir
            if not run(f'uv venv{venv_args} {vdir}'):
                utils.rm_vdir(vdir, args)
                return f'Error: failed to recreate {vdir} for {pkgname}.'

            if not utils.piprun(vdir, f'sync{pip_args} --reinstall {tfile}'):
                utils.rm_vdir(vdir, args)
                return f'Error: failed to resync {pkgname}'

        err = utils.make_links(vdir, pkgname, args, data)
        if err:
            return err

        print(f'{pkgname} reinstalled.')

    return None
